import numpy as np

from typing import Any
from typing import Dict
from typing import Tuple
from typing import Optional
from dataclasses import dataclass
from cftool.misc import print_warning

from .file import FileParserBlock
from ....schema import DataTypes
from ....schema import DataBundle
from ....schema import ColumnTypes
from ....schema import INoInitDataBlock


TRes = Tuple[ColumnTypes, int]


@dataclass
class MLRecognizerConfig:
    all_close_threshold: float = 1.0e-6
    redundancy_threshold: float = 0.5
    custom_feature_types: Optional[Dict[str, ColumnTypes]] = None


@INoInitDataBlock.register("ml_recognizer")
class RecognizerBlock(INoInitDataBlock):
    config: MLRecognizerConfig  # type: ignore
    feature_types: Dict[str, ColumnTypes]
    label_types: Dict[str, ColumnTypes]
    num_unique_features: Dict[str, int]
    num_unique_labels: Dict[str, int]
    # map from original indices to new indices (which exclude the REDUNDANT columns)
    index_mapping: Dict[str, int]

    # inheritance

    def to_info(self) -> Dict[str, Any]:
        return dict(
            feature_types=self.feature_types,
            label_types=self.label_types,
            num_unique_features=self.num_unique_features,
            num_unique_labels=self.num_unique_labels,
            index_mapping=self.index_mapping,
        )

    def transform(self, bundle: DataBundle, for_inference: bool) -> DataBundle:
        if not isinstance(bundle.x_train, np.ndarray):
            raise ValueError("`RecognizerBlock` should be used on numpy features")
        target = [
            int(str_idx)
            for str_idx, column_type in self.feature_types.items()
            if column_type != ColumnTypes.REDUNDANT
        ]
        bundle.x_train = bundle.x_train[..., target]
        if bundle.x_valid is not None:
            if not isinstance(bundle.x_valid, np.ndarray):
                raise ValueError("`RecognizerBlock` should be used on numpy features")
            bundle.x_valid = bundle.x_valid[..., target]
        return bundle

    def fit_transform(self, bundle: DataBundle) -> DataBundle:
        x_train: np.ndarray = bundle.x_train
        y_train: np.ndarray = bundle.y_train
        self.feature_types = {}
        self.label_types = {}
        self.num_unique_features = {}
        self.num_unique_labels = {}
        for idx, line in enumerate(x_train.T):
            self._fit(idx, line, False, self.config.custom_feature_types)
        for idx, line in enumerate(y_train.T):
            self._fit(idx, line, True, None)
        counter = 0
        self.index_mapping = {}
        for i in range(x_train.shape[-1]):
            str_i = str(i)
            if self.feature_types[str_i] == ColumnTypes.REDUNDANT:
                continue
            self.index_mapping[str_i] = counter
            counter += 1
        return self.transform(bundle, False)

    # api

    @property
    def file_parser(self) -> Optional[FileParserBlock]:
        return self.try_get_previous(FileParserBlock)

    # internal

    def _fit(
        self,
        idx: int,
        line: np.ndarray,
        is_label: bool,
        custom_types: Optional[Dict[str, ColumnTypes]],
    ) -> None:
        str_idx = str(idx)
        file_parser = self.file_parser
        if file_parser is None or file_parser.is_placeholder:
            int_line = line.astype(int).astype(np.float64)
            dtype = DataTypes.INT if np.allclose(line, int_line) else DataTypes.FLOAT
            key = name = str_idx
        else:
            attr = f"{'label' if is_label else 'feature'}_header"
            header = getattr(file_parser, attr)
            dtype = file_parser.converters[header[idx]].dtype
            key = header[idx]
            name = f"'{key}'"
        custom_type = None if custom_types is None else custom_types.get(key)
        column_type, num_unique = self._fit_with(name, line, dtype, custom_type)
        (self.label_types if is_label else self.feature_types)[str_idx] = column_type
        if column_type != ColumnTypes.NUMERICAL:
            uniques = self.num_unique_labels if is_label else self.num_unique_features
            uniques[str_idx] = num_unique

    def _fit_with(
        self,
        name: str,
        line: np.ndarray,
        dtype: DataTypes,
        custom_type: Optional[ColumnTypes],
    ) -> TRes:
        prefix = f"values in {dtype} column {name}"
        postfix = "It'll be marked as redundant."
        if dtype == DataTypes.FLOAT:
            threshold = self.config.all_close_threshold
            if line.max().item() - line.min().item() <= threshold:
                if self.is_local_rank_0:
                    print_warning(f"{prefix} are ALL CLOSE. {postfix}")
                return ColumnTypes.REDUNDANT, 0
            if np.all(line == 0.0):
                if self.is_local_rank_0:
                    print_warning(f"{prefix} are ALL ZERO. {postfix}")
                return ColumnTypes.REDUNDANT, 0
            if custom_type is None:
                return ColumnTypes.NUMERICAL, 0
            return custom_type, 0
        num_unique = len(np.unique(line))
        num_samples = len(line)
        if num_unique == 1:
            if self.is_local_rank_0:
                print_warning(f"{prefix} are ALL SAME. {postfix}")
            return ColumnTypes.REDUNDANT, num_unique
        if num_unique == num_samples:
            if self.is_local_rank_0:
                print_warning(f"{prefix} are ALL DIFFERENT. {postfix}")
            return ColumnTypes.REDUNDANT, num_unique
        unique_ratio = num_unique / num_samples
        if unique_ratio >= self.config.redundancy_threshold:
            if self.is_local_rank_0:
                msg = f"{prefix} are TOO MANY (ratio={unique_ratio:8.6f}). {postfix}"
                print_warning(msg)
            return ColumnTypes.REDUNDANT, num_unique
        if custom_type is None:
            return ColumnTypes.CATEGORICAL, num_unique
        return custom_type, num_unique


__all__ = [
    "ColumnTypes",
    "MLRecognizerConfig",
    "RecognizerBlock",
]



from typing import Any
from typing import Dict
from typing import List
from typing import Type
from typing import Optional
from collections import OrderedDict
from cftool.misc import safe_execute
from cftool.misc import print_warning
from cftool.misc import shallow_copy_dict
from cftool.pipeline import IBlock
from cftool.pipeline import IPipeline
from cftool.pipeline import TPipeline

from ..schema import IData
from ..schema import DLConfig
from ..misc.toolkit import get_ddp_info


class Block(IBlock):
    """
    Building block of a `Pipeline`.

    * data: only available in `TrainingPipeline`.
    * training_workspace: only available in `TrainingPipeline`, identical to `config.workspace`.
    * serialize_folder: only available in `load` process.
    * previous: previous building blocks in `build` process. Will be ALL building blocks in `run` process.

    """

    data: Optional[IData]
    training_workspace: Optional[str]
    serialize_folder: Optional[str]
    previous: Dict[str, "Block"]

    # optional callbacks

    def process_defaults(self, _defaults: OrderedDict) -> None:
        pass

    def run(self, data: IData, _defaults: OrderedDict, **kwargs: Any) -> None:
        pass

    def save_extra(self, folder: str) -> None:
        pass

    def load_from(self, folder: str) -> None:
        pass

    # api

    @property
    def ddp(self) -> bool:
        return get_ddp_info() is not None

    @property
    def local_rank(self) -> Optional[int]:
        ddp_info = get_ddp_info()
        return None if ddp_info is None else ddp_info.local_rank

    @property
    def is_local_rank_0(self) -> bool:
        return not self.ddp or self.local_rank == 0


class Pipeline(IPipeline):
    data: Optional[IData] = None
    training_workspace: Optional[str] = None
    serialize_folder: Optional[str] = None
    config: DLConfig
    blocks: List[Block]
    _defaults: OrderedDict
    config_file = "config.json"

    # inheritance

    @classmethod
    def init(cls: Type[TPipeline], config: DLConfig) -> TPipeline:
        config.sanity_check()
        self: Pipeline = cls()
        self.config = config.copy()
        self._defaults = OrderedDict()
        return self

    @property
    def config_base(self) -> Type[DLConfig]:
        return DLConfig

    @property
    def block_base(self) -> Type[Block]:
        return Block

    def to_info(self) -> Dict[str, Any]:
        info = super().to_info()
        info["_defaults"] = [[k, v] for k, v in self._defaults.items()]
        return info

    def from_info(self, info: Dict[str, Any]) -> None:
        self._defaults = OrderedDict()
        for k, v in info["_defaults"]:
            self._defaults[k] = v
        super().from_info(info)

    def before_block_build(self, block: Block) -> None:
        block.data = self.data
        block.training_workspace = self.training_workspace
        if self.serialize_folder is None:
            block.serialize_folder = None
        else:
            block.serialize_folder = self.serialize_folder

    def after_block_build(self, block: Block) -> None:
        block.process_defaults(self._defaults)
        if self.training_workspace is not None:
            if self.training_workspace != self.config.workspace:
                self.training_workspace = self.config.workspace

    # api

    def run(self, data: IData, **kwargs: Any) -> None:
        if not self.blocks:
            print_warning("no blocks are built, nothing will happen")
            return
        kw = shallow_copy_dict(kwargs)
        kw["data"] = data
        kw["_defaults"] = self._defaults
        all_blocks = self.block_mappings
        for block in self.blocks:
            block.previous = all_blocks
            safe_execute(block.run, kw)


__all__ = [
    "Pipeline",
]



import torch

from torch import Tensor
from typing import Any
from typing import Dict
from typing import Optional
from typing import NamedTuple

from .encoders import Encoder
from .encoders import EncodingResult
from ...schema import IMLModel
from ...schema import MLEncoderSettings
from ...schema import MLGlobalEncoderSettings


class MLEncodePack(NamedTuple):
    one_hot: Optional[Tensor]
    embedding: Optional[Tensor]
    numerical: Optional[Tensor]
    merged_categorical: Optional[Tensor]
    merged_all: Tensor


class MLModel(IMLModel):
    def __init__(
        self,
        *args: Any,
        encoder_settings: Optional[Dict[str, MLEncoderSettings]] = None,
        global_encoder_settings: Optional[MLGlobalEncoderSettings] = None,
        **kwargs: Any
    ):
        super().__init__()
        if encoder_settings is None:
            self.encoder = None
        else:
            self.encoder = Encoder(encoder_settings, global_encoder_settings)

    def encode(self, net: Tensor) -> MLEncodePack:
        if self.encoder is None or self.encoder.is_empty:
            return MLEncodePack(None, None, net, None, net)
        numerical_columns = [
            index
            for index in range(net.shape[-1])
            if index not in self.encoder.tgt_columns
        ]
        numerical = net[..., numerical_columns]
        res: EncodingResult = self.encoder(net)
        merged_categorical = res.merged
        if merged_categorical is None:
            merged_all = numerical
        else:
            merged_all = torch.cat([numerical, merged_categorical], dim=-1)
        return MLEncodePack(
            res.one_hot,
            res.embedding,
            numerical,
            merged_categorical,
            merged_all,
        )


__all__ = [
    "MLModel",
]



from typing import Any
from typing import Dict
from typing import Tuple

from ....schema import IRuntimeDataBlock
from ....constants import INPUT_KEY
from ....constants import LABEL_KEY
from ....constants import ORIGINAL_LABEL_KEY


@IRuntimeDataBlock.register("tuple_to_batch")
class TupleToBatchBlock(IRuntimeDataBlock):
    def postprocess_item(self, item: Tuple[Any, Any]) -> Dict[str, Any]:
        image, labels = item
        return {
            INPUT_KEY: image,
            LABEL_KEY: labels,
            ORIGINAL_LABEL_KEY: labels,
        }


__all__ = ["TupleToBatchBlock"]



import torch

import torch.nn as nn

from collections import OrderedDict


def make_layers(block, no_relu_layers):
    layers = []
    for layer_name, v in block.items():
        if "pool" in layer_name:
            layer = nn.MaxPool2d(kernel_size=v[0], stride=v[1], padding=v[2])
            layers.append((layer_name, layer))
        else:
            conv2d = nn.Conv2d(
                in_channels=v[0],
                out_channels=v[1],
                kernel_size=v[2],
                stride=v[3],
                padding=v[4],
            )
            layers.append((layer_name, conv2d))
            if layer_name not in no_relu_layers:
                layers.append(("relu_" + layer_name, nn.ReLU(inplace=True)))

    return nn.Sequential(OrderedDict(layers))


class bodypose_model(nn.Module):
    def __init__(self):
        super(bodypose_model, self).__init__()

        # these layers have no relu layer
        no_relu_layers = [
            "conv5_5_CPM_L1",
            "conv5_5_CPM_L2",
            "Mconv7_stage2_L1",
            "Mconv7_stage2_L2",
            "Mconv7_stage3_L1",
            "Mconv7_stage3_L2",
            "Mconv7_stage4_L1",
            "Mconv7_stage4_L2",
            "Mconv7_stage5_L1",
            "Mconv7_stage5_L2",
            "Mconv7_stage6_L1",
            "Mconv7_stage6_L1",
        ]
        blocks = {}
        block0 = OrderedDict(
            [
                ("conv1_1", [3, 64, 3, 1, 1]),
                ("conv1_2", [64, 64, 3, 1, 1]),
                ("pool1_stage1", [2, 2, 0]),
                ("conv2_1", [64, 128, 3, 1, 1]),
                ("conv2_2", [128, 128, 3, 1, 1]),
                ("pool2_stage1", [2, 2, 0]),
                ("conv3_1", [128, 256, 3, 1, 1]),
                ("conv3_2", [256, 256, 3, 1, 1]),
                ("conv3_3", [256, 256, 3, 1, 1]),
                ("conv3_4", [256, 256, 3, 1, 1]),
                ("pool3_stage1", [2, 2, 0]),
                ("conv4_1", [256, 512, 3, 1, 1]),
                ("conv4_2", [512, 512, 3, 1, 1]),
                ("conv4_3_CPM", [512, 256, 3, 1, 1]),
                ("conv4_4_CPM", [256, 128, 3, 1, 1]),
            ]
        )

        # Stage 1
        block1_1 = OrderedDict(
            [
                ("conv5_1_CPM_L1", [128, 128, 3, 1, 1]),
                ("conv5_2_CPM_L1", [128, 128, 3, 1, 1]),
                ("conv5_3_CPM_L1", [128, 128, 3, 1, 1]),
                ("conv5_4_CPM_L1", [128, 512, 1, 1, 0]),
                ("conv5_5_CPM_L1", [512, 38, 1, 1, 0]),
            ]
        )

        block1_2 = OrderedDict(
            [
                ("conv5_1_CPM_L2", [128, 128, 3, 1, 1]),
                ("conv5_2_CPM_L2", [128, 128, 3, 1, 1]),
                ("conv5_3_CPM_L2", [128, 128, 3, 1, 1]),
                ("conv5_4_CPM_L2", [128, 512, 1, 1, 0]),
                ("conv5_5_CPM_L2", [512, 19, 1, 1, 0]),
            ]
        )
        blocks["block1_1"] = block1_1
        blocks["block1_2"] = block1_2

        self.model0 = make_layers(block0, no_relu_layers)

        # Stages 2 - 6
        for i in range(2, 7):
            blocks["block%d_1" % i] = OrderedDict(
                [
                    ("Mconv1_stage%d_L1" % i, [185, 128, 7, 1, 3]),
                    ("Mconv2_stage%d_L1" % i, [128, 128, 7, 1, 3]),
                    ("Mconv3_stage%d_L1" % i, [128, 128, 7, 1, 3]),
                    ("Mconv4_stage%d_L1" % i, [128, 128, 7, 1, 3]),
                    ("Mconv5_stage%d_L1" % i, [128, 128, 7, 1, 3]),
                    ("Mconv6_stage%d_L1" % i, [128, 128, 1, 1, 0]),
                    ("Mconv7_stage%d_L1" % i, [128, 38, 1, 1, 0]),
                ]
            )

            blocks["block%d_2" % i] = OrderedDict(
                [
                    ("Mconv1_stage%d_L2" % i, [185, 128, 7, 1, 3]),
                    ("Mconv2_stage%d_L2" % i, [128, 128, 7, 1, 3]),
                    ("Mconv3_stage%d_L2" % i, [128, 128, 7, 1, 3]),
                    ("Mconv4_stage%d_L2" % i, [128, 128, 7, 1, 3]),
                    ("Mconv5_stage%d_L2" % i, [128, 128, 7, 1, 3]),
                    ("Mconv6_stage%d_L2" % i, [128, 128, 1, 1, 0]),
                    ("Mconv7_stage%d_L2" % i, [128, 19, 1, 1, 0]),
                ]
            )

        for k in blocks.keys():
            blocks[k] = make_layers(blocks[k], no_relu_layers)

        self.model1_1 = blocks["block1_1"]
        self.model2_1 = blocks["block2_1"]
        self.model3_1 = blocks["block3_1"]
        self.model4_1 = blocks["block4_1"]
        self.model5_1 = blocks["block5_1"]
        self.model6_1 = blocks["block6_1"]

        self.model1_2 = blocks["block1_2"]
        self.model2_2 = blocks["block2_2"]
        self.model3_2 = blocks["block3_2"]
        self.model4_2 = blocks["block4_2"]
        self.model5_2 = blocks["block5_2"]
        self.model6_2 = blocks["block6_2"]

    def forward(self, x):
        out1 = self.model0(x)

        out1_1 = self.model1_1(out1)
        out1_2 = self.model1_2(out1)
        out2 = torch.cat([out1_1, out1_2, out1], 1)

        out2_1 = self.model2_1(out2)
        out2_2 = self.model2_2(out2)
        out3 = torch.cat([out2_1, out2_2, out1], 1)

        out3_1 = self.model3_1(out3)
        out3_2 = self.model3_2(out3)
        out4 = torch.cat([out3_1, out3_2, out1], 1)

        out4_1 = self.model4_1(out4)
        out4_2 = self.model4_2(out4)
        out5 = torch.cat([out4_1, out4_2, out1], 1)

        out5_1 = self.model5_1(out5)
        out5_2 = self.model5_2(out5)
        out6 = torch.cat([out5_1, out5_2, out1], 1)

        out6_1 = self.model6_1(out6)
        out6_2 = self.model6_2(out6)

        return out6_1, out6_2


class handpose_model(nn.Module):
    def __init__(self):
        super(handpose_model, self).__init__()

        # these layers have no relu layer
        no_relu_layers = [
            "conv6_2_CPM",
            "Mconv7_stage2",
            "Mconv7_stage3",
            "Mconv7_stage4",
            "Mconv7_stage5",
            "Mconv7_stage6",
        ]
        # stage 1
        block1_0 = OrderedDict(
            [
                ("conv1_1", [3, 64, 3, 1, 1]),
                ("conv1_2", [64, 64, 3, 1, 1]),
                ("pool1_stage1", [2, 2, 0]),
                ("conv2_1", [64, 128, 3, 1, 1]),
                ("conv2_2", [128, 128, 3, 1, 1]),
                ("pool2_stage1", [2, 2, 0]),
                ("conv3_1", [128, 256, 3, 1, 1]),
                ("conv3_2", [256, 256, 3, 1, 1]),
                ("conv3_3", [256, 256, 3, 1, 1]),
                ("conv3_4", [256, 256, 3, 1, 1]),
                ("pool3_stage1", [2, 2, 0]),
                ("conv4_1", [256, 512, 3, 1, 1]),
                ("conv4_2", [512, 512, 3, 1, 1]),
                ("conv4_3", [512, 512, 3, 1, 1]),
                ("conv4_4", [512, 512, 3, 1, 1]),
                ("conv5_1", [512, 512, 3, 1, 1]),
                ("conv5_2", [512, 512, 3, 1, 1]),
                ("conv5_3_CPM", [512, 128, 3, 1, 1]),
            ]
        )

        block1_1 = OrderedDict(
            [("conv6_1_CPM", [128, 512, 1, 1, 0]), ("conv6_2_CPM", [512, 22, 1, 1, 0])]
        )

        blocks = {}
        blocks["block1_0"] = block1_0
        blocks["block1_1"] = block1_1

        # stage 2-6
        for i in range(2, 7):
            blocks["block%d" % i] = OrderedDict(
                [
                    ("Mconv1_stage%d" % i, [150, 128, 7, 1, 3]),
                    ("Mconv2_stage%d" % i, [128, 128, 7, 1, 3]),
                    ("Mconv3_stage%d" % i, [128, 128, 7, 1, 3]),
                    ("Mconv4_stage%d" % i, [128, 128, 7, 1, 3]),
                    ("Mconv5_stage%d" % i, [128, 128, 7, 1, 3]),
                    ("Mconv6_stage%d" % i, [128, 128, 1, 1, 0]),
                    ("Mconv7_stage%d" % i, [128, 22, 1, 1, 0]),
                ]
            )

        for k in blocks.keys():
            blocks[k] = make_layers(blocks[k], no_relu_layers)

        self.model1_0 = blocks["block1_0"]
        self.model1_1 = blocks["block1_1"]
        self.model2 = blocks["block2"]
        self.model3 = blocks["block3"]
        self.model4 = blocks["block4"]
        self.model5 = blocks["block5"]
        self.model6 = blocks["block6"]

    def forward(self, x):
        out1_0 = self.model1_0(x)
        out1_1 = self.model1_1(out1_0)
        concat_stage2 = torch.cat([out1_1, out1_0], 1)
        out_stage2 = self.model2(concat_stage2)
        concat_stage3 = torch.cat([out_stage2, out1_0], 1)
        out_stage3 = self.model3(concat_stage3)
        concat_stage4 = torch.cat([out_stage3, out1_0], 1)
        out_stage4 = self.model4(concat_stage4)
        concat_stage5 = torch.cat([out_stage4, out1_0], 1)
        out_stage5 = self.model5(concat_stage5)
        concat_stage6 = torch.cat([out_stage5, out1_0], 1)
        out_stage6 = self.model6(concat_stage6)
        return out_stage6



import os
import dill
import json
import random
import shutil

import numpy as np

from abc import abstractmethod
from PIL import Image
from enum import Enum
from tqdm import tqdm
from typing import Any
from typing import Set
from typing import Dict
from typing import List
from typing import Type
from typing import Tuple
from typing import Union
from typing import Callable
from typing import Optional
from typing import NamedTuple
from cftool.cv import to_rgb
from cftool.dist import Parallel
from cftool.misc import walk
from cftool.misc import hash_dict
from cftool.misc import is_numeric
from cftool.misc import print_info
from cftool.misc import print_error
from cftool.misc import print_warning
from cftool.misc import get_arguments
from cftool.misc import shallow_copy_dict
from cftool.misc import ISerializable

from ....schema import IDataset
from ....schema import IDataBlock
from ....schema import DataBundle
from ....constants import INPUT_KEY
from ....constants import LABEL_KEY

try:
    import lmdb
except:
    lmdb = None


READY_FILE = "READY"
preparations: Dict[str, Type["IPreparation"]] = {}
default_image_extensions = {".jpg", ".png", ".jpeg"}


class IPreparation(ISerializable):
    d = preparations

    # abstract

    @abstractmethod
    def filter(self, hierarchy: List[str]) -> bool:
        """return whether the `hierarchy` should be abandoned"""

    @abstractmethod
    def get_label(self, hierarchy: List[str]) -> Any:
        """
        return the label of the `hierarchy`
        * notice that the returned label should be hashable
        """

    @abstractmethod
    def copy(self, src_path: str, tgt_path: str) -> None:
        """
        copy an image from `src_path` to `tgt_path`.
        * If any error occurred in this method, we will handle it and filter out the
        corresponding sample, so feel free to do preprocessing or do sanity checks
        here and raise errors!
        """

    # optional callbacks

    @property
    def extra_labels(self) -> Optional[List[str]]:
        """
        overwrite this property if extra labels are needed.
        * each of the returned element will be passed to the `get_extra_label` method
        below as the `label_name` argument.
        """
        return None

    def prepare_src_folder(self, src_folder: str) -> None:
        """
        prepare the source folder before other processes happen.
        * common use case is to 'augment' the source folder beforehand.
        """

    def get_extra_label(self, label_name: str, hierarchy: List[str]) -> Any:
        """
        return extra label.
        * will not be called if the above `extra_labels` property returns `None`.
        """

    def to_info(self) -> Dict[str, Any]:
        return {}

    def from_info(self, info: Dict[str, Any]) -> None:
        """
        `IPreparation` will not be created from scratch, and will always be
        instantiated with the `from_pack` method. So if we need to assign
        some attributes, we need to do them here.
        * See `ResizedPreparation` below as an example.
        """

    # api

    def is_ready(
        self,
        tgt_folder: str,
        valid_split: Union[int, float],
        folder_hash: str,
    ) -> bool:
        candidates = [DatasetSplit.TRAIN]
        if valid_split > 0:
            candidates.append(DatasetSplit.VALID)
        for split in candidates:
            split_folder = os.path.join(tgt_folder, split)
            extra_keys = [f"{key}_{LABEL_KEY}" for key in self.extra_labels or []]
            for key in [LABEL_KEY] + extra_keys:
                path = os.path.join(split_folder, f"{key}.json")
                if not os.path.isfile(path):
                    return False
            ready_path = os.path.join(split_folder, READY_FILE)
            if not os.path.isfile(ready_path):
                return False
            with open(ready_path, "r") as f:
                if f.read().strip() != folder_hash:
                    return False
        return True

    def get_num_classes(self, tgt_folder: str) -> Dict[str, int]:
        num_classes = {}
        for label_name in [LABEL_KEY] + (self.extra_labels or []):
            path = os.path.join(tgt_folder, f"idx2{label_name}.json")
            if not os.path.isfile(path):
                num_classes[label_name] = 0
                continue
            with open(path, "r") as f:
                num_classes[label_name] = len(json.load(f))
        return num_classes


@IPreparation.register("default")
class DefaultPreparation(IPreparation):
    """
    A default preparation.
    * All images are treated as valid.
    * No labels will be calculated.
    """

    def filter(self, hierarchy: List[str]) -> bool:
        return True

    def get_label(self, hierarchy: List[str]) -> Any:
        return 0

    def copy(self, src_path: str, tgt_path: str) -> None:
        shutil.copyfile(src_path, tgt_path)


@IPreparation.register("resized")
class ResizedPreparation(DefaultPreparation):
    """
    A Preparation with images resized.
    * Useful when we need a static-sized image dataset.
    * The attribute, `img_size`, should be initialized when `from_info` is called.
    """

    img_size: int
    keep_aspect_ratio: bool

    def copy(self, src_path: str, tgt_path: str) -> None:
        img = to_rgb(Image.open(src_path))
        if not self.keep_aspect_ratio:
            img.resize((self.img_size, self.img_size), Image.LANCZOS).save(tgt_path)
            return None
        w, h = img.size
        wh_ratio = w / h
        if wh_ratio < 1:
            new_w = self.img_size
            new_h = round(self.img_size / wh_ratio)
        else:
            new_w = round(self.img_size * wh_ratio)
            new_h = self.img_size
        img.resize((new_w, new_h), Image.LANCZOS).save(tgt_path)

    def to_info(self) -> Dict[str, Any]:
        return dict(img_size=self.img_size, keep_aspect_ratio=self.keep_aspect_ratio)

    def from_info(self, info: Dict[str, Any]) -> None:
        img_size = info.get("img_size")
        if img_size is None:
            raise ValueError("`img_size` should be provided for `ResizedPreparation`")
        self.img_size = img_size
        self.keep_aspect_ratio = info.setdefault("keep_aspect_ratio", True)


def default_lmdb_path(folder: str, split: str) -> str:
    return f"{folder}.{split}"


class LMDBItem(NamedTuple):
    image: np.ndarray
    labels: Dict[str, Any]


def prepare_image_folder(
    *,
    src_folder: str,
    tgt_folder: str,
    to_index: bool,
    prefix: Optional[str],
    preparation_pack: Optional[Dict[str, Any]],
    force_rerun: bool,
    extensions: Optional[Set[str]],
    make_labels_in_parallel: bool,
    save_data_in_parallel: bool,
    num_jobs: int,
    train_all_data: bool,
    valid_split: Union[int, float],
    max_num_valid: int,
    lmdb_config: Optional[Dict[str, Any]],
    use_tqdm: bool,
    strict: bool,
) -> Tuple[str, Optional[List[str]]]:
    f"""
    Efficiently convert an arbitrary image folder (`src_folder`) to
    an image folder (`tgt_folder`) which can be parsed by `carefree-learn`.
    * It offers various customization options such as parallel processing,
    validation split, and LMDB integration.

    Parameters
    ----------
    src_folder : str
        The path to the source folder containing the images to be processed.
    tgt_folder : str
        The path to the target folder where the processed images and related metadata will be saved.
    to_index : bool
        Whether to convert the labels to indices or not.
    prefix : Optional[str], default: None
        An optional prefix for the source and target folder paths.
    preparation_pack : Optional[Dict[str, Any]], default: None
        An optional JSON configuration for customizing the image preparation process.
        * should be of format: dict(type="...", info=dict(...))
    force_rerun : bool
        If True, the function will rerun even if the target folder already exists and is ready.
    extensions : Optional[Set[str]], default: None
        A set of allowed image file extensions. Defaults to {default_image_extensions} if not provided.
    make_labels_in_parallel : bool
        Whether to create labels in parallel or not.
    save_data_in_parallel : bool
        Whether to save data in parallel or not.
    num_jobs : int
        The number of jobs to use for parallel processing.
    train_all_data : bool
        If True, all data will be used for training, and no validation set will be created.
    valid_split : Union[int, float]
        The number or proportion of samples to be used for validation. If a float is provided,
        it represents the proportion of the dataset to include in the validation set.
    max_num_valid : int
        The maximum number of samples to be used for validation.
    lmdb_config : Optional[Dict[str, Any]], default: None
        An optional dictionary containing LMDB configuration. If not provided, LMDB will not be used.
    use_tqdm : bool
        If True, display progress bars using `tqdm`.
    strict : bool
        If True, error will be raised if any invalid sample occurred.

    Returns
    -------
    str
        The path to the target folder containing the processed images and related metadata.
    """

    if prefix is not None:
        src_folder = os.path.join(prefix, src_folder)
        tgt_folder = os.path.join(prefix, tgt_folder)
    if preparation_pack is None:
        preparation_pack = dict(type="default", info={})
    preparation: IPreparation = IPreparation.from_pack(preparation_pack)
    preparation.prepare_src_folder(src_folder)

    if train_all_data:
        valid_split = 0
    folder_hash = hash_dict(
        dict(
            src_folder=src_folder,
            to_index=to_index,
            preparation_pack=preparation_pack,
            extensions=extensions,
            train_all_data=train_all_data,
            valid_split=valid_split,
            max_num_valid=max_num_valid,
            lmdb_config=lmdb_config,
            strict=strict,
        )
    )
    if not force_rerun and preparation.is_ready(tgt_folder, valid_split, folder_hash):
        return tgt_folder, preparation.extra_labels

    if os.path.isdir(tgt_folder):
        print_warning(f"'{tgt_folder}' already exists, it will be removed")
        shutil.rmtree(tgt_folder)
    os.makedirs(tgt_folder, exist_ok=True)

    print_info("collecting hierarchies")

    def hierarchy_callback(hierarchy: List[str], path: str) -> None:
        hierarchy = hierarchy[prefix_idx:]
        if not preparation.filter(hierarchy):
            return None
        hierarchy_list.append(hierarchy)
        all_img_paths.append(path)

    all_img_paths: List[str] = []
    hierarchy_list: List[List[str]] = []
    if extensions is None:
        extensions = default_image_extensions
    prefix_idx = 0
    if prefix is not None:
        prefix_idx = len(prefix.split(os.path.sep))
    walk(src_folder, hierarchy_callback, extensions)

    def get_labels(
        label_fn: Callable,
        label_name: Optional[str] = None,
    ) -> List[Any]:
        def task(h: List[str]) -> Any:
            try:
                args = (h,) if label_name is None else (label_name, h)
                return label_fn(*args)
            except Exception as err:
                err_path = "/".join(h)
                print_error(f"error occurred ({err}) when getting label of {err_path}")
                return None

        if not make_labels_in_parallel:
            return [task(h) for h in tqdm(hierarchy_list)]
        parallel = Parallel(num_jobs, use_tqdm=use_tqdm)
        num_files = len(hierarchy_list)
        random_indices = np.random.permutation(num_files).tolist()
        shuffled = [hierarchy_list[i] for i in random_indices]
        groups = parallel.grouped(task, shuffled).ordered_results
        shuffled_results: List[Any] = sum(groups, [])
        final_results = [None] * num_files
        for idx, rs in zip(random_indices, shuffled_results):
            final_results[idx] = rs
        return final_results

    print_info("making labels")
    labels = get_labels(preparation.get_label)
    excluded_indices = {i for i, label in enumerate(labels) if label is None}
    extra_labels_dict: Optional[Dict[str, List[str]]] = None
    extra_labels = preparation.extra_labels
    extra_label_fn = preparation.get_extra_label
    if extra_labels is not None:
        extra_labels_dict = {}
        print_info("making extra labels")
        for el_name in extra_labels:
            extra_labels_dict[el_name] = get_labels(extra_label_fn, el_name)
        for extra_labels in extra_labels_dict.values():
            for i, extra_label in enumerate(extra_labels):
                if extra_label is None:
                    excluded_indices.add(i)

    # exclude samples
    if excluded_indices:
        if not strict:
            print_warning(f"{len(excluded_indices)} samples will be excluded")
        else:
            raise ValueError(
                "\n".join(
                    [
                        "following samples are invalid:",
                        *[f"* {all_img_paths[i]}" for i in excluded_indices],
                        "please check the log history for more details",
                    ]
                )
            )
    for i in sorted(excluded_indices)[::-1]:
        labels.pop(i)
        if extra_labels_dict is not None:
            for sub_labels in extra_labels_dict.values():
                sub_labels.pop(i)
        all_img_paths.pop(i)

    def get_raw_2idx(raw_labels: List[Any]) -> Dict[Any, Any]:
        return {
            v: numpy_token if isinstance(v, str) and v.endswith(".npy") else v
            for v in raw_labels
        }

    def check_dump_mappings(l2i: Dict[Any, Any]) -> bool:
        all_indices = set(l2i.values())
        if len(all_indices) > 1:
            return True
        return list(all_indices)[0] != numpy_token

    numpy_token = "[NUMPY]"
    if to_index:
        label2idx = {label: i for i, label in enumerate(sorted(set(labels)))}
        labels_dict = {"": [label2idx[label] for label in labels]}
        dump_mappings = True
    else:
        labels_dict = {"": labels}
        label2idx = get_raw_2idx(sorted(set(labels)))
        dump_mappings = check_dump_mappings(label2idx)

    open_file_from = lambda folder: lambda file: open(
        os.path.join(folder, file), "w", encoding="utf-8"
    )
    open_tgt_file = open_file_from(tgt_folder)

    if dump_mappings:
        with open_tgt_file(f"{LABEL_KEY}2idx.json") as f:
            json.dump(label2idx, f, ensure_ascii=False)
        with open_tgt_file(f"idx2{LABEL_KEY}.json") as f:
            json.dump({v: k for k, v in label2idx.items()}, f, ensure_ascii=False)

    if extra_labels_dict is not None:
        for el_name, label_collection in extra_labels_dict.items():
            if not to_index:
                labels_dict[el_name] = label_collection  # type: ignore
                extra2idx = get_raw_2idx(sorted(set(label_collection)))
                dump_mappings = check_dump_mappings(extra2idx)
            else:
                extra2idx = {
                    extra_label: i  # type: ignore
                    for i, extra_label in enumerate(sorted(set(label_collection)))
                }
                labels_dict[el_name] = [extra2idx[el] for el in label_collection]
                dump_mappings = True
            if dump_mappings:
                with open_tgt_file(f"{el_name}2idx.json") as f:
                    json.dump(extra2idx, f, ensure_ascii=False)
                with open_tgt_file(f"idx2{el_name}.json") as f:
                    eld = {v: k for k, v in extra2idx.items()}
                    json.dump(eld, f, ensure_ascii=False)

    # prepare core
    def save(indices: np.ndarray, d_num_jobs: int, dtype: str) -> None:
        def record(idx: int) -> Optional[Tuple[str, Dict[str, Any]]]:
            split_folder = os.path.join(tgt_folder, dtype)
            os.makedirs(split_folder, exist_ok=True)
            img_path = all_img_paths[idx]
            ext = os.path.splitext(img_path)[1]
            new_img_path = os.path.join(split_folder, f"{idx}{ext}")
            try:
                preparation.copy(img_path, new_img_path)
                key = os.path.abspath(new_img_path)
                idx_labels: Dict[str, Any] = {}
                for label_t, t_labels in labels_dict.items():
                    idx_labels[label_t] = {key: t_labels[idx]}
                return key, idx_labels
            except Exception as err:
                print_error(f"error occurred with {img_path} : {err}")
                return None

        print_info(f"saving {dtype} dataset")
        results: List[Tuple[str, Dict[str, Any]]]
        indices = indices.copy()
        np.random.shuffle(indices)
        if not save_data_in_parallel:
            results = [record(i) for i in indices]  # type: ignore
        else:
            parallel = Parallel(d_num_jobs, use_tqdm=use_tqdm)
            results = sum(parallel.grouped(record, indices).ordered_results, [])
        d_valid_indices = [i for i, r in enumerate(results) if r is not None]
        results = [results[i] for i in d_valid_indices]
        valid_paths = [all_img_paths[idx] for idx in indices[d_valid_indices]]
        new_paths, all_labels_list = zip(*results)
        merged_labels = shallow_copy_dict(all_labels_list[0])
        for sub_labels_ in all_labels_list[1:]:
            for k, v in shallow_copy_dict(sub_labels_).items():
                merged_labels[k].update(v)
        print_info(
            "\n".join(
                [
                    "",
                    "=" * 100,
                    f"num {dtype} samples : {len(next(iter(merged_labels.values())))}",
                    f"num {dtype} label types : {len(merged_labels)}",
                    "-" * 100,
                    "",
                ]
            )
        )
        open_dtype_file = open_file_from(os.path.join(tgt_folder, dtype))

        with open_dtype_file("paths.json") as f_:
            json.dump(new_paths, f_, ensure_ascii=False)
        path_mapping = dict(zip(new_paths, valid_paths))
        with open_dtype_file("path_mapping.json") as f_:
            json.dump(path_mapping, f_, ensure_ascii=False)
        for label_type, type_labels in merged_labels.items():
            delim = "_" if label_type else ""
            label_file = f"{label_type}{delim}{LABEL_KEY}.json"
            with open_dtype_file(label_file) as f_:
                json.dump(type_labels, f_, ensure_ascii=False)
        # lmdb
        if lmdb_config is None or lmdb is None:
            if lmdb_config is not None:
                msg = "`lmdb` is not installed, so `lmdb_config` will be ignored"
                print_warning(msg)
        else:
            local_lmdb_config = shallow_copy_dict(lmdb_config)
            local_lmdb_config.setdefault("path", default_lmdb_path(tgt_folder, dtype))
            local_lmdb_config.setdefault("map_size", 1099511627776 * 2)
            db = lmdb.open(**local_lmdb_config)
            context = db.begin(write=True)
            d_num_samples = len(results)
            iterator = zip(range(d_num_samples), new_paths, all_labels_list)
            if use_tqdm:
                iterator = tqdm(iterator, total=d_num_samples, desc="lmdb")
            for i, path, i_labels in iterator:
                i_new_labels = {}
                for k, v in i_labels.items():
                    vv = v[path]
                    if isinstance(vv, str):
                        if vv.endswith(".npy"):
                            vv = np.load(vv)
                    i_new_labels[k] = vv
                context.put(
                    str(i).encode("ascii"),
                    dill.dumps(LMDBItem(np.array(Image.open(path)), i_new_labels)),
                )
            context.put(
                "length".encode("ascii"),
                str(d_num_samples).encode("ascii"),
            )
            context.commit()
            db.sync()
            db.close()
        # dump READY
        with open_dtype_file(READY_FILE) as f_:
            f_.write(folder_hash)

    num_sample = len(all_img_paths)
    if isinstance(valid_split, float):
        if valid_split < 0.0 or valid_split >= 1.0:
            raise ValueError("`valid_split` should be within [0, 1)")
        valid_split = max(1, min(max_num_valid, int(round(num_sample * valid_split))))
    assert isinstance(valid_split, int)

    if valid_split <= 0:
        save(np.arange(num_sample), max(1, num_jobs), DatasetSplit.TRAIN)
        return tgt_folder, extra_labels

    train_portion = (num_sample - valid_split) / num_sample
    label_indices_mapping: Dict[Any, List[int]] = {}
    for i, label in enumerate(labels):
        if isinstance(label, str) and label.endswith(".npy"):
            label = numpy_token
        label_indices_mapping.setdefault(label, []).append(i)
    tuple(map(random.shuffle, label_indices_mapping.values()))
    train_indices_list: List[List[int]] = []
    valid_indices_list: List[List[int]] = []
    for label_indices in label_indices_mapping.values():
        num_label_samples = len(label_indices)
        num_train = int(round(train_portion * num_label_samples))
        num_train = min(num_train, num_label_samples - 1)
        if num_train == 0:
            train_indices_list.append([label_indices[0]])
        else:
            train_indices_list.append(label_indices[:num_train])
        valid_indices_list.append(label_indices[num_train:])

    def propagate(src: List[List[int]], tgt: List[List[int]]) -> None:
        resolved = 0
        src_lengths = list(map(len, src))
        sorted_indices = np.argsort(src_lengths).tolist()[::-1]
        while True:
            for idx in sorted_indices:
                if len(src[idx]) > 1:
                    tgt[idx].append(src[idx].pop())
                resolved += 1
                if resolved == diff:
                    break
            if resolved == diff:
                break

    diff = sum(map(len, valid_indices_list)) - valid_split
    if diff > 0:
        propagate(valid_indices_list, train_indices_list)
    elif diff < 0:
        diff *= -1
        propagate(train_indices_list, valid_indices_list)
    merged_train_indices: List[int] = sorted(set(sum(train_indices_list, [])))
    merged_valid_indices: List[int] = sorted(set(sum(valid_indices_list, [])))
    train_indices = np.array(merged_train_indices)
    valid_indices = np.array(merged_valid_indices)

    save(train_indices, max(1, num_jobs), DatasetSplit.TRAIN)
    save(valid_indices, max(1, num_jobs // 2), DatasetSplit.VALID)
    return tgt_folder, extra_labels


class DatasetSplit(str, Enum):
    TRAIN = "train"
    VALID = "valid"


class ImagePaths(NamedTuple):
    length: int
    img_paths: List[str]
    # label_name -> path (from `img_paths`) -> label (of the corresponding path)
    labels: Optional[Dict[str, Dict[str, Any]]]
    lmdb_context: Optional[Any]

    def to_dataset(self) -> "ImageFolderDataset":
        return ImageFolderDataset(self)


def get_image_paths(
    split: DatasetSplit,
    base_folder: str,
    extra_label_names: Optional[List[str]],
    lmdb_config: Optional[Dict[str, Any]],
) -> Optional[ImagePaths]:
    folder = os.path.abspath(os.path.join(base_folder, split))
    if not os.path.isdir(folder):
        return None
    support_extensions = default_image_extensions
    support_extensions.add(".npy")
    img_paths = list(
        map(
            lambda file: os.path.join(folder, file),
            filter(
                lambda file: os.path.splitext(file)[1] in support_extensions,
                os.listdir(folder),
            ),
        )
    )
    if lmdb_config is None or lmdb is None:
        if lmdb_config is not None:
            print_warning("`lmdb` is not installed, so `lmdb_config` will be ignored")
        lmdb_context = None
        with open(os.path.join(folder, f"{LABEL_KEY}.json"), "r") as f:
            labels = {LABEL_KEY: json.load(f)}
        if extra_label_names is not None:
            for name in extra_label_names:
                el_file = f"{name}_{LABEL_KEY}.json"
                with open(os.path.join(folder, el_file), "r") as f:
                    labels[name] = json.load(f)
        length = len(img_paths)
    else:
        new_lmdb_config = shallow_copy_dict(lmdb_config)
        new_lmdb_config.setdefault("path", default_lmdb_path(base_folder, split))
        new_lmdb_config.setdefault("lock", False)
        new_lmdb_config.setdefault("meminit", False)
        new_lmdb_config.setdefault("readonly", True)
        new_lmdb_config.setdefault("readahead", False)
        new_lmdb_config.setdefault("max_readers", 32)
        lmdb_ = lmdb.open(**new_lmdb_config)
        lmdb_context = lmdb_.begin(buffers=True, write=False)
        with lmdb_.begin(write=False) as context:
            length = int(context.get("length".encode("ascii")).decode("ascii"))
    return ImagePaths(length, img_paths, labels, lmdb_context)


class ImageFolderDataset(IDataset):
    def __init__(self, paths: ImagePaths) -> None:
        self.paths = paths

    def __len__(self) -> int:
        return self.paths.length

    def __getitem__(self, item: Union[int, List[int], np.ndarray]) -> Dict[str, Any]:
        if not isinstance(item, int):
            raise ValueError("`ImageFolderDataset` only supports single index")
        if self.paths.lmdb_context is not None:
            key = str(item).encode("ascii")
            lmdb_item: LMDBItem = dill.loads(self.paths.lmdb_context.get(key))
            new_item = lmdb_item.labels
            for k, v in new_item.items():
                if is_numeric(v):
                    new_item[k] = np.array([v])
            new_item[INPUT_KEY] = lmdb_item.image
            return new_item
        img_path = self.paths.img_paths[item]
        image = np.array(Image.open(img_path))
        if self.paths.labels is None:
            return {INPUT_KEY: image}
        new_item = {k: v[img_path] for k, v in self.paths.labels.items()}
        for k, v in new_item.items():
            if is_numeric(v):
                v = np.array([v])
            elif isinstance(v, str) and v.endswith(".npy"):
                v = np.load(v)
            new_item[k] = v
        new_item[INPUT_KEY] = image
        return new_item


@IDataBlock.register("image_folder")
class ImageFolderBlock(IDataBlock):
    """
    This block will utilize `prepare_image_folder` method to create an image folder
    which can be parsed by `carefree-learn`.

    It will then transfer the input `DataBundle`, which should only contain `x_train` indicating
    the original image folder, to a new `DataBundle` containing the `ImageFolderDataset`(s).

    TODO: Support parsing `x_valid` and use it directly as validation `ImageFolderDataset`.
    """

    tgt_folder: Optional[str]
    to_index: bool
    prefix: Optional[str]
    preparation_pack: Optional[Dict[str, Any]]
    force_rerun: bool
    extensions: Optional[Set[str]]
    make_labels_in_parallel: bool
    save_data_in_parallel: bool
    num_jobs: int
    train_all_data: bool
    valid_split: Union[int, float]
    max_num_valid: int
    lmdb_config: Optional[Dict[str, Any]]
    use_tqdm: bool
    strict: bool

    def __init__(
        self,
        *,
        tgt_folder: Optional[str] = None,
        to_index: bool = False,
        prefix: Optional[str] = None,
        preparation_pack: Optional[Dict[str, Any]] = None,
        force_rerun: bool = False,
        extensions: Optional[Set[str]] = None,
        make_labels_in_parallel: bool = False,
        save_data_in_parallel: bool = True,
        num_jobs: int = 8,
        train_all_data: bool = False,
        valid_split: Union[int, float] = 0.1,
        max_num_valid: int = 10000,
        lmdb_config: Optional[Dict[str, Any]] = None,
        use_tqdm: bool = True,
        strict: bool = False,
    ) -> None:
        super().__init__(**get_arguments())

    # inherit

    @property
    def fields(self) -> List[str]:
        return [
            "tgt_folder",
            "to_index",
            "prefix",
            "preparation_pack",
            "force_rerun",
            "extensions",
            "make_labels_in_parallel",
            "save_data_in_parallel",
            "num_jobs",
            "train_all_data",
            "valid_split",
            "max_num_valid",
            "lmdb_config",
            "use_tqdm",
            "strict",
        ]

    def transform(self, bundle: DataBundle, for_inference: bool) -> DataBundle:
        train_paths, valid_paths = self._parse(bundle.x_train, for_inference)  # type: ignore
        bundle.x_train = train_paths.to_dataset()
        if valid_paths is not None:
            bundle.x_valid = valid_paths.to_dataset()
        return bundle

    def fit_transform(self, bundle: DataBundle) -> DataBundle:
        return self.transform(bundle, False)

    # internal

    def _parse(
        self,
        src_folder: str,
        for_inference: bool,
    ) -> Tuple[ImagePaths, Optional[ImagePaths]]:
        if src_folder is None:
            return None
        info = shallow_copy_dict(self.to_info())
        # get tgt folder
        if self.tgt_folder is None:
            tgt_folder = f"{src_folder}_{'inference' if for_inference else 'training'}"
        else:
            if not for_inference:
                tgt_folder = self.tgt_folder
            else:
                tgt_folder = f"{self.tgt_folder}_inference"
        # set properties
        info["src_folder"] = src_folder
        info["tgt_folder"] = tgt_folder
        if for_inference:
            info["train_all_data"] = True
        # prepare
        tgt_folder, extra_labels = prepare_image_folder(**info)
        # get paths
        common_args = tgt_folder, extra_labels, self.lmdb_config
        train_paths = get_image_paths(DatasetSplit.TRAIN, *common_args)
        valid_paths = get_image_paths(DatasetSplit.VALID, *common_args)
        return train_paths, valid_paths  # type: ignore


__all__ = [
    "default_image_extensions",
    "IPreparation",
    "DefaultPreparation",
    "ResizedPreparation",
    "ImageFolderBlock",
]



import torch

from numpy import ndarray
from torch import Tensor
from typing import Tuple
from collections import namedtuple

try:
    import cv2
except:
    cv2 = None


APack = Tuple[ndarray, ndarray]
TPack = Tuple[Tensor, Tensor]


class EvalTransform:
    def __init__(self):
        pass

    def transform(self, image: Tensor, mask: Tensor) -> TPack:
        raise NotImplementedError

    def inv_transform(self, image: Tensor) -> Tensor:
        raise NotImplementedError


class PadToDivisor(EvalTransform):
    """
    Pad side of the image so that its side is divisible by divisor.

    Args:
        divisor (int): desirable image size divisor
        border_mode (OpenCV flag): OpenCV border mode.
        fill_value (int, float, list of int, lisft of float): padding value if border_mode is cv2.BORDER_CONSTANT.
    """

    PadParams = namedtuple("PadParams", ["top", "bottom", "left", "right"])

    def __init__(
        self,
        divisor: int,
        border_mode: int = 0 if cv2 is None else cv2.BORDER_CONSTANT,
        fill_value: int = 0,
    ):
        super().__init__()
        self.border_mode = border_mode
        self.fill_value = fill_value
        self.divisor = divisor
        self._pads = None

    def transform(self, image: ndarray, mask: ndarray) -> APack:
        if cv2 is None:
            raise ValueError("`cv2` is needed for `PadToDivisor`")

        self._pads = PadToDivisor.PadParams(
            *self._get_dim_padding(image.shape[0]),
            *self._get_dim_padding(image.shape[1]),
        )

        image = cv2.copyMakeBorder(
            image,
            *self._pads,
            self.border_mode,
            value=self.fill_value,
        )
        mask = cv2.copyMakeBorder(
            mask,
            *self._pads,
            self.border_mode,
            value=self.fill_value,
        )
        return image, mask

    def inv_transform(self, image: Tensor) -> Tensor:
        assert (
            self._pads is not None
        ), "Something went wrong, inv_transform(...) should be called after transform(...)"
        return self._remove_padding(image)

    def _get_dim_padding(self, dim_size: int) -> Tuple[int, int]:
        pad = (self.divisor - dim_size % self.divisor) % self.divisor
        pad_upper = pad // 2
        pad_lower = pad - pad_upper
        return pad_upper, pad_lower

    def _remove_padding(self, tensor: Tensor) -> Tensor:
        tensor_h, tensor_w = tensor.shape[:2]
        cropped = tensor[
            self._pads.top : tensor_h - self._pads.bottom,
            self._pads.left : tensor_w - self._pads.right,
            :,
        ]
        return cropped


class NormalizeTensor(EvalTransform):
    def __init__(self, mean: float, std: float, device: torch.device):
        super().__init__()
        self.mean = torch.as_tensor(mean).reshape(1, 3, 1, 1).to(device)
        self.std = torch.as_tensor(std).reshape(1, 3, 1, 1).to(device)

    def transform(self, image: Tensor, mask: Tensor) -> TPack:
        image.sub_(self.mean).div_(self.std)
        return image, mask

    def inv_transform(self, image: Tensor) -> Tensor:
        image.mul_(self.std).add_(self.mean)
        return image


class ToTensor(EvalTransform):
    def __init__(self, device: torch.device):
        super().__init__()
        self.device = device

    def transform(self, image: ndarray, mask: ndarray) -> TPack:
        image = torch.as_tensor(image, device=self.device, dtype=torch.float32)
        mask = torch.as_tensor(mask, device=self.device)
        image.unsqueeze_(0)
        mask.unsqueeze_(0).unsqueeze_(0)
        return image.permute(0, 3, 1, 2) / 255.0, mask

    def inv_transform(self, image: Tensor) -> Tensor:
        image.squeeze_(0)
        return 255 * image.permute(1, 2, 0)


class AddFlippedTensor(EvalTransform):
    def transform(self, image: Tensor, mask: Tensor) -> TPack:
        flipped_image = torch.flip(image, dims=(3,))
        flipped_mask = torch.flip(mask, dims=(3,))
        image = torch.cat((image, flipped_image), dim=0)
        mask = torch.cat((mask, flipped_mask), dim=0)
        return image, mask

    def inv_transform(self, image: Tensor) -> Tensor:
        return 0.5 * (image[:1] + torch.flip(image[1:], dims=(3,)))



import torch
import random

from abc import abstractmethod
from abc import ABCMeta
from torch import Tensor
from typing import Any
from typing import Optional
from cftool.misc import safe_execute
from cftool.misc import shallow_copy_dict
from cftool.types import tensor_dict_type

from ...schema import _forward
from ...constants import INPUT_KEY
from ...constants import LABEL_KEY
from ...constants import PREDICTIONS_KEY
from ...misc.toolkit import slerp
from ...misc.toolkit import get_device


class GeneratorMixin(metaclass=ABCMeta):
    latent_dim: int
    num_classes: Optional[int] = None

    # inherit

    @property
    @abstractmethod
    def can_reconstruct(self) -> bool:
        pass

    @abstractmethod
    def sample(
        self,
        num_samples: int,
        *,
        class_idx: Optional[int] = None,
        **kwargs: Any,
    ) -> Tensor:
        pass

    @abstractmethod
    def interpolate(
        self,
        num_samples: int,
        *,
        class_idx: Optional[int] = None,
        use_slerp: bool = False,
        **kwargs: Any,
    ) -> Tensor:
        pass

    # utilities

    @property
    def is_conditional(self) -> bool:
        return self.num_classes is not None

    def reconstruct(self, net: Tensor, **kwargs: Any) -> Tensor:
        name = self.__class__.__name__
        if not self.can_reconstruct:
            raise ValueError(f"`{name}` does not support `reconstruct`")
        labels = kwargs.pop(LABEL_KEY, None)
        batch = {INPUT_KEY: net, LABEL_KEY: labels}
        if self.is_conditional and labels is None:
            raise ValueError(
                f"`{LABEL_KEY}` should be provided in `reconstruct` "
                f"for conditional `{name}`"
            )
        return _forward(self, 0, batch, INPUT_KEY, **kwargs)[PREDICTIONS_KEY]

    def get_sample_labels(
        self,
        num_samples: int,
        class_idx: Optional[int] = None,
    ) -> Optional[Tensor]:
        if self.num_classes is None:
            return None
        if class_idx is not None:
            return torch.full([num_samples], class_idx, device=get_device(self))
        return torch.randint(self.num_classes, [num_samples], device=get_device(self))


class GaussianGeneratorMixin(GeneratorMixin, metaclass=ABCMeta):
    @abstractmethod
    def decode(self, z: Tensor, *, labels: Optional[Tensor], **kwargs: Any) -> Tensor:
        pass

    def generate_z(self, num_samples: int) -> Tensor:
        return torch.randn(num_samples, self.latent_dim, device=get_device(self))

    def sample(
        self,
        num_samples: int,
        *,
        class_idx: Optional[int] = None,
        labels: Optional[Tensor] = None,
        **kwargs: Any,
    ) -> Tensor:
        z = self.generate_z(num_samples)
        if labels is None:
            labels = self.get_sample_labels(num_samples, class_idx)
        elif class_idx is not None:
            msg = "`class_idx` should not be provided when `labels` is provided"
            raise ValueError(msg)
        kw = shallow_copy_dict(kwargs)
        kw["z"] = z
        kw["labels"] = labels
        return safe_execute(self.decode, kw)

    def interpolate(
        self,
        num_samples: int,
        *,
        class_idx: Optional[int] = None,
        use_slerp: bool = False,
        **kwargs: Any,
    ) -> Tensor:
        z1 = self.generate_z(1)
        z2 = self.generate_z(1)
        shape = z1.shape
        z1 = z1.view(1, -1)
        z2 = z2.view(1, -1)
        ratio = torch.linspace(0.0, 1.0, num_samples, device=get_device(self))[:, None]
        z = slerp(z1, z2, ratio) if use_slerp else ratio * z1 + (1.0 - ratio) * z2
        z = z.view(num_samples, *shape[1:])
        if class_idx is None and self.num_classes is not None:
            class_idx = random.randint(0, self.num_classes - 1)
        kw = shallow_copy_dict(kwargs)
        kw["z"] = z
        kw["labels"] = self.get_sample_labels(num_samples, class_idx)
        return safe_execute(self.decode, kw)


class ImageTranslatorMixin:
    def onnx_forward(self, batch: tensor_dict_type) -> Any:
        return self.generate_from(batch[INPUT_KEY], determinate=True)

    def generate_from(self, net: Tensor, **kwargs: Any) -> Tensor:
        rs = _forward(self, 0, {INPUT_KEY: net}, INPUT_KEY, **kwargs)[PREDICTIONS_KEY]
        if isinstance(rs, list):
            rs = rs[0]
        return rs


__all__ = [
    "GeneratorMixin",
    "GaussianGeneratorMixin",
    "ImageTranslatorMixin",
]



import torch

import torch.nn as nn

from torch import Tensor
from typing import Any
from typing import Dict
from typing import Type
from typing import Optional
from cftool.misc import safe_execute
from cftool.misc import check_requires
from cftool.misc import WithRegister
from cftool.types import tensor_dict_type

from ....schema import _forward
from ....schema import TrainerState
from ....constants import INPUT_KEY
from ....constants import LATENT_KEY
from ....misc.toolkit import eval_context
from ....misc.toolkit import get_device
from ....modules.blocks import ImgToPatches


encoders: Dict[str, Type["EncoderMixin"]] = {}
encoders_1d: Dict[str, Type["Encoder1DMixin"]] = {}


class IEncoder:
    def encode(self, batch: tensor_dict_type, **kwargs: Any) -> Tensor:
        return run_encoder(self, 0, batch, **kwargs)[LATENT_KEY]


# encode to a latent feature map
class EncoderMixin(IEncoder, WithRegister["EncoderMixin"]):
    d = encoders

    in_channels: int
    num_downsample: int
    latent_channels: int

    def latent_resolution(self, img_size: int) -> int:
        shape = 1, self.in_channels, img_size, img_size
        with eval_context(self):
            net = self.encode({INPUT_KEY: torch.zeros(*shape, device=get_device(self))})
        return net.shape[2]


# encode to a 1d latent code
class Encoder1DMixin(IEncoder, WithRegister["Encoder1DMixin"]):
    d = encoders_1d

    in_channels: int
    latent_dim: int


def make_encoder(name: str, config: Dict[str, Any], *, is_1d: bool = False) -> IEncoder:
    base = (Encoder1DMixin if is_1d else EncoderMixin).get(name)  # type: ignore
    return safe_execute(base, config)


def run_encoder(
    encoder: IEncoder,
    batch_idx: int,
    batch: tensor_dict_type,
    state: Optional[TrainerState] = None,
    **kwargs: Any,
) -> tensor_dict_type:
    return _forward(
        encoder,
        batch_idx,
        batch,
        INPUT_KEY,
        state,
        general_output_key=LATENT_KEY,
        **kwargs,
    )


# from patches
class EncoderFromPatchesMixin:
    encoder: nn.Module

    def __init__(
        self,
        *,
        img_size: int,
        patch_size: int,
        in_channels: int,
        latent_dim: int = 128,
        to_patches_type: str = "vanilla",
        to_patches_config: Optional[Dict[str, Any]] = None,
    ):
        if to_patches_config is None:
            to_patches_config = {}
        to_patches_config.update(
            {
                "img_size": img_size,
                "patch_size": patch_size,
                "in_channels": in_channels,
                "latent_dim": latent_dim,
            }
        )
        self.to_patches = ImgToPatches.make(to_patches_type, to_patches_config)

    @property
    def num_patches(self) -> int:
        return self.to_patches.num_patches

    def forward(self, net: Tensor, **kwargs: Any) -> Tensor:
        determinate = kwargs.pop("determinate", False)
        patches, hw = self.to_patches(net, determinate=determinate)
        kwargs["hw"] = hw
        kwargs["hwp"] = *net.shape[-2:], self.to_patches.patch_size
        if check_requires(self.encoder.forward, "determinate", strict=False):
            kwargs["determinate"] = determinate
        return self.encoder(patches, **kwargs)


class Encoder1DFromPatches(EncoderFromPatchesMixin, Encoder1DMixin, nn.Module):
    def __init__(
        self,
        *,
        img_size: int,
        patch_size: int,
        in_channels: int,
        latent_dim: int = 128,
        to_patches_type: str = "vanilla",
        to_patches_config: Optional[Dict[str, Any]] = None,
    ):
        nn.Module.__init__(self)
        EncoderFromPatchesMixin.__init__(
            self,
            img_size=img_size,
            patch_size=patch_size,
            in_channels=in_channels,
            latent_dim=latent_dim,
            to_patches_type=to_patches_type,
            to_patches_config=to_patches_config,
        )
        self.in_channels = in_channels
        self.latent_dim = latent_dim


class Encoder2DFromPatches(EncoderFromPatchesMixin, EncoderMixin, nn.Module):
    def __init__(
        self,
        *,
        img_size: int,
        patch_size: int,
        in_channels: int,
        latent_channels: int = 128,
        to_patches_type: str = "vanilla",
        to_patches_config: Optional[Dict[str, Any]] = None,
    ):
        nn.Module.__init__(self)
        EncoderFromPatchesMixin.__init__(
            self,
            img_size=img_size,
            patch_size=patch_size,
            in_channels=in_channels,
            latent_dim=latent_channels,
            to_patches_type=to_patches_type,
            to_patches_config=to_patches_config,
        )
        self.in_channels = in_channels
        self.latent_channels = latent_channels


__all__ = [
    "make_encoder",
    "run_encoder",
    "IEncoder",
    "EncoderMixin",
    "Encoder1DMixin",
    "Encoder1DFromPatches",
    "Encoder2DFromPatches",
]



import torch

from torch import nn
from torch import Tensor
from typing import Any
from typing import Set
from typing import Dict
from typing import List
from typing import Union
from typing import Generic
from typing import TypeVar
from typing import Optional
from typing import NamedTuple
from cftool.misc import print_info
from cftool.misc import print_warning
from cftool.misc import shallow_copy_dict
from cftool.array import tensor_dict_type

from .hooks import IHook
from .hooks import IBasicHook
from .hooks import IAttentionHook
from .hooks import MultiHooks
from .customs import Linear


class IHijackMixin:
    hook: Optional[IHook]


# linear/conv hijacks


class IBasicHijackMixin(IHijackMixin):
    weight: Tensor

    def __init__(self, *args: Any, hook: Optional[IBasicHook] = None, **kwargs: Any):
        self.args = args
        self.kwargs = shallow_copy_dict(kwargs)
        super().__init__(*args, **kwargs)
        self.hook = hook

    def forward(self, net: Tensor) -> Tensor:
        inp = net
        if self.hook is not None:
            inp = self.hook.before_forward(inp)
        net = super().forward(net)  # type: ignore
        if self.hook is not None:
            net = self.hook.after_forward(inp, net)
        return net


class HijackLinear(IBasicHijackMixin, nn.Linear):
    pass


class HijackCustomLinear(IBasicHijackMixin, Linear):
    pass


class HijackConv1d(IBasicHijackMixin, nn.Conv1d):
    pass


class HijackConv2d(IBasicHijackMixin, nn.Conv2d):
    pass


class HijackConv3d(IBasicHijackMixin, nn.Conv3d):
    pass


# attention hijacks


class IAttention:
    in_w: Optional[nn.Parameter]
    hook: Optional[IAttentionHook]
    input_dim: int
    embed_dim: int


# LoRA
## input -> lora_down -> selector -> lora_up -> dropout -> alpha (scale)


TLoRA = TypeVar("TLoRA")


class ILoRAHook(Generic[TLoRA]):
    _ms: List[TLoRA]
    backup: Optional[Tensor] = None
    injected: bool = False

    @property
    def m(self) -> TLoRA:
        return self._ms[0]

    def inject(self, w: nn.Parameter, updown: Tensor, index: Optional[int]) -> None:
        if self.backup is None and (index is None or index == 0):
            self.backup = w.data
        if self.backup is None:
            w.data = w.data + updown
        else:
            w.data = self.backup + updown
        self.injected = True


class ILoRAMappingHook(ILoRAHook[Union[nn.Linear, nn.Conv2d]], IBasicHook):
    rank: int
    lora_down: Union[nn.Linear, nn.Conv2d]
    selector: Union[nn.Identity, nn.Linear, nn.Conv2d]
    lora_up: Union[nn.Linear, nn.Conv2d]
    dropout: nn.Dropout
    alpha: nn.Parameter
    scale: float = 1.0

    @property
    def alpha_scale(self) -> float:
        return self.alpha.item() / self.lora_down.weight.shape[0]

    def set_scale(self, scale: float) -> None:
        self.scale = scale
        self.injected = False

    def get_updown(self) -> Tensor:
        up = self.lora_up.weight
        down = self.lora_down.weight
        if len(up.shape) == 2 and len(down.shape) == 2:
            updown = up @ down
        else:
            updown = (up[..., 0, 0] @ down[..., 0, 0])[..., None, None]
        updown = self.scale * self.alpha_scale * updown
        return updown

    def before_forward(self, inp: Tensor, index: Optional[int] = None) -> Tensor:
        if not self.injected:
            weight = self.m.weight
            updown = self.get_updown()
            self.inject(weight, updown, index)
        return inp

    @classmethod
    def create_with(cls, m: IBasicHijackMixin, rank: int) -> "ILoRAMappingHook":
        self = cls(*m.args, rank=rank, **m.kwargs)
        self.to(m.weight)
        self._ms = [m]
        return self


class LoRALinearHook(ILoRAMappingHook):
    def __init__(
        self,
        in_features: int,
        out_features: int,
        *args: Any,
        rank: int = 4,
        dropout: float = 0.1,
        scale: float = 0.1,
        **kwargs: Any,
    ) -> None:
        super().__init__()
        if rank > min(in_features, out_features):
            raise ValueError(
                f"LoRA rank {rank} must be less or equal than "
                f"{min(in_features, out_features)}"
            )
        self.rank = rank
        self.lora_down = nn.Linear(in_features, rank, bias=False)
        self.selector = nn.Identity()
        self.lora_up = nn.Linear(rank, out_features, bias=False)
        self.dropout = nn.Dropout(dropout)
        self.register_parameter("alpha", nn.Parameter(torch.tensor(scale)))

        nn.init.normal_(self.lora_down.weight, std=1.0 / rank)
        nn.init.zeros_(self.lora_up.weight)

    def set_selector(self, diag: Tensor) -> None:
        assert diag.shape == (self.rank,)
        self.selector = nn.Linear(self.rank, self.rank, bias=False)
        self.selector.weight.data = torch.diag(diag).to(self.lora_up.weight.data)


class LoRAConvHook(ILoRAMappingHook):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        stride: int = 1,
        padding: int = 0,
        dilation: int = 1,
        groups: int = 1,
        *args: Any,
        rank: int = 4,
        dropout: float = 0.1,
        scale: float = 0.1,
        **kwargs: Any,
    ) -> None:
        super().__init__()
        if rank > min(in_channels, out_channels):
            raise ValueError(
                f"LoRA rank {rank} must be less or equal than "
                f"{min(in_channels, out_channels)}"
            )
        self.rank = rank
        self.lora_down = nn.Conv2d(
            in_channels=in_channels,
            out_channels=rank,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            groups=groups,
            bias=False,
        )
        self.selector = nn.Identity()
        self.lora_up = nn.Conv2d(
            in_channels=rank,
            out_channels=out_channels,
            kernel_size=1,
            stride=1,
            padding=0,
            bias=False,
        )
        self.dropout = nn.Dropout(dropout)
        self.register_parameter("alpha", nn.Parameter(torch.tensor(scale)))

        nn.init.normal_(self.lora_down.weight, std=1.0 / rank)
        nn.init.zeros_(self.lora_up.weight)

    def set_selector(self, diag: Tensor) -> None:
        assert diag.shape == (self.rank,)
        self.selector = nn.Conv2d(
            in_channels=self.rank,
            out_channels=self.rank,
            kernel_size=1,
            stride=1,
            padding=0,
            bias=False,
        )
        self.selector.weight.data = torch.diag(diag).to(self.lora_up.weight.data)


class LoRAAttentionHook(ILoRAHook[IAttention], IAttentionHook):
    def __init__(self, in_features: int, out_features: int, *, rank: int) -> None:
        super().__init__()
        self.to_q = LoRALinearHook(in_features, out_features, rank=rank)
        self.to_k = LoRALinearHook(in_features, out_features, rank=rank)
        self.to_v = LoRALinearHook(in_features, out_features, rank=rank)

    def set_scale(self, scale: float) -> None:
        self.to_q.set_scale(scale)
        self.to_k.set_scale(scale)
        self.to_v.set_scale(scale)
        self.injected = False

    def before_forward(self, inp: Tensor, index: Optional[int] = None) -> Tensor:
        if not self.injected:
            in_w = self.m.in_w
            hooks = [self.to_q, self.to_k, self.to_v]
            updown = torch.vstack([h.get_updown() for h in hooks])
            assert in_w is not None, "should be self_attn in lora"
            self.inject(in_w, updown, index)
        return inp

    @classmethod
    def create_with(cls, m: IAttention, rank: int) -> "LoRAAttentionHook":
        self = cls(m.input_dim, m.embed_dim, rank=rank)
        self._ms = [m]
        return self


class LoRAPack(NamedTuple):
    rank: int
    hooks: Dict[str, Union[ILoRAMappingHook, LoRAAttentionHook]]


class LoRAManager:
    def __init__(self) -> None:
        self.m = None
        self.injected = False
        self.lora_packs: Dict[str, LoRAPack] = {}

    # api

    def build_pack(self, *keys: str) -> Optional[LoRAPack]:
        if not keys:
            return None
        if len(keys) == 1:
            return self.lora_packs.get(keys[0])
        selected = [self.lora_packs.get(key) for key in keys]
        not_none = [pack for pack in selected if pack is not None]
        if not not_none:
            return None
        if len(not_none) == 1:
            return selected[0]
        selected_ranks = list(set(pack.rank for pack in not_none))
        rank = -1 if len(selected_ranks) > 1 else selected_ranks[0]
        hooks: Dict[str, List[IHook]] = {}
        for pack in not_none:
            for k, v in pack.hooks.items():
                hooks.setdefault(k, []).append(v)
        merged = {k: v[0] if len(v) == 1 else MultiHooks(v) for k, v in hooks.items()}
        return LoRAPack(rank, merged)

    def load_pack_with(self, key: str, d: tensor_dict_type) -> None:
        pack = self.lora_packs.get(key)
        if pack is None:
            raise ValueError(f"cannot find '{key}' pack")
        num_source = len(d)
        num_prepared = 0
        for k, v in pack.hooks.items():
            for n, p in v.named_parameters():
                tp = d[f"{k}.{n}"]
                assert p.data.shape == tp.shape
                p.data = tp.to(p.data)
                num_prepared += 1
        if num_source == num_prepared:
            print_info(f"{num_prepared} weights are loaded")
        else:
            print_warning(
                f"only {num_prepared}/{num_source} weights are loaded, "
                "which could lead to unexpected behaviours"
            )

    def prepare(
        self,
        m: nn.Module,
        *,
        key: str,
        rank: int,
        target_ancestors: Optional[Set[str]] = None,
    ) -> None:
        # cleanup existing hooks
        pack = self.lora_packs.pop(key, None)
        if pack is not None:
            for hook in pack.hooks.values():
                hook.cpu()
            torch.cuda.empty_cache()
        # prepare new hooks
        pack = LoRAPack(rank, {})
        self.m = m
        if target_ancestors is None:
            self._prepare(m, "", pack)
        else:
            for name, module in m.named_modules():
                if module.__class__.__name__ in target_ancestors:
                    self._prepare(module, name, pack)
        # set pack
        self.lora_packs[key] = pack

    def inject(self, m: nn.Module, *keys: str) -> None:
        if self.m is None:
            raise ValueError("LoRA is not prepared, please call `prepare` first.")
        if m is not self.m:
            raise ValueError("prepared module does not match the incoming module.")
        pack = self.build_pack(*keys)
        if pack is None:
            raise ValueError(f"cannot build LoRAPack with {', '.join(keys)}")
        hooks = pack.hooks
        pivot = list(m.parameters())[0]
        num_hooks = len(hooks)
        num_injected = 0
        for name, module in m.named_modules():
            hook = hooks.get(name)
            if hook is not None:
                if not isinstance(module, IHijackMixin):
                    msg = f"`hook` is found for '{name}' but it is not a Hijack module"
                    raise ValueError(msg)
                hook.to(pivot)
                module.hook = hook
                num_injected += 1
        if num_hooks == num_injected:
            print_info(f"{num_injected} hooks are injected")
        else:
            print_warning(
                f"only {num_injected}/{num_hooks} hooks are injected, "
                "which could lead to unexpected behaviours"
            )
        self.injected = True

    def cleanup(self, m: nn.Module) -> None:
        if not self.injected:
            raise ValueError("LoRA is not injected, please call `inject` first.")
        if m is not self.m:
            raise ValueError("injected module does not match the incoming module.")
        for module in m.modules():  # type: ignore
            if isinstance(module, IHijackMixin):
                hook = module.hook
                if not isinstance(hook, MultiHooks):
                    hooks = [hook]
                else:
                    hooks = hook.hooks
                for h in hooks:
                    if not isinstance(h, ILoRAHook):
                        continue
                    h.injected = False
                    module.hook = None
                    if h.backup is not None:
                        if isinstance(module, IAttention):
                            module.in_w.data = h.backup  # type: ignore
                        else:
                            module.weight.data = h.backup  # type: ignore
                        h.backup = None
        self.injected = False
        torch.cuda.empty_cache()

    def set_scale(self, key: str, scale: float) -> None:
        pack = self.lora_packs.get(key)
        if pack is None:
            raise ValueError(f"cannot find LoRAPack '{key}'")
        for hook in pack.hooks.values():
            hook.set_scale(scale)

    def set_scales(self, scales: Dict[str, float]) -> None:
        for k, v in scales.items():
            self.set_scale(k, v)

    def checkpoint(self, m: nn.Module) -> Dict[str, Optional[IHook]]:
        hooks: Dict[str, Optional[IHook]] = {}
        for key, module in m.named_modules():
            if isinstance(module, IHijackMixin):
                hooks[key] = module.hook
        return hooks

    def restore(self, m: nn.Module, hooks: Dict[str, Optional[IHook]]) -> None:
        if m is not self.m:
            raise ValueError("prepared module does not match the incoming module.")
        for key, module in m.named_modules():  # type: ignore
            if isinstance(module, IHijackMixin):
                module.hook = hooks[key]
        self.injected = True

    # internal

    def _prepare(self, m: nn.Module, prefix: str, pack: LoRAPack) -> None:
        rank = pack.rank
        for name, module in m.named_modules():
            key = f"{prefix}.{name}"
            if isinstance(module, (HijackLinear, HijackCustomLinear)):
                pack.hooks[key] = LoRALinearHook.create_with(module, rank)
            elif isinstance(module, HijackConv2d):
                pack.hooks[key] = LoRAConvHook.create_with(module, rank)
            elif isinstance(module, IAttention):
                pack.hooks[key] = LoRAAttentionHook.create_with(module, rank)



import torch

import numpy as np

from typing import Any
from typing import Dict
from typing import List
from typing import Type
from typing import Tuple
from typing import Union
from typing import Optional
from cftool.array import to_numpy
from cftool.types import np_dict_type
from torch.utils.data import Sampler
from torch.utils.data import DataLoader
from torch.utils.data import SequentialSampler
from torch.utils.data.dataloader import _BaseDataLoaderIter
from torch.utils.data.distributed import DistributedSampler

from ...schema import IData
from ...schema import IDataset
from ...schema import DataConfig
from ...schema import IDataLoader
from ...schema import DataProcessor
from ...schema import DataProcessorConfig
from ...misc.toolkit import get_ddp_info
from ...misc.toolkit import get_world_size


class TorchDataset(IDataset):
    def __init__(self, dataset: IDataset, processor: DataProcessor) -> None:
        self.dataset = dataset
        self.processor = processor

    def __len__(self) -> int:
        return len(self.dataset)

    def __getitem__(self, item: Union[int, List[int], np.ndarray]) -> np_dict_type:
        batch = self.dataset[item]
        batch = self.processor.postprocess_item(batch)
        return batch


class _DataLoader(DataLoader):
    def __init__(
        self,
        dataset: TorchDataset,
        batch_size: Optional[int] = 1,
        shuffle: bool = False,
        sampler: Optional[Sampler[int]] = None,
        *,
        use_distributed_sampler: Optional[bool] = None,
        **kwargs: Any,
    ):
        if use_distributed_sampler is None:
            use_distributed_sampler = get_ddp_info() is not None
        if use_distributed_sampler:
            if sampler is not None and not isinstance(sampler, DistributedSampler):
                raise ValueError(
                    "`sampler` should be `DistributedSampler` "
                    "when `use_distributed_sampler` is True"
                )
            sampler = DistributedSampler(dataset, shuffle=shuffle)
            shuffle = False
        super().__init__(dataset, batch_size, shuffle, sampler, **kwargs)

    def __setattr__(self, attr: str, val: Any) -> None:
        if self.__initialized and attr in (
            "batch_size",
            "batch_sampler",
            "drop_last",
            "dataset",
            "persistent_workers",
        ):
            raise ValueError(
                f"{attr} attribute should not be set after "
                f"{self.__class__.__name__} is initialized"
            )

        super(DataLoader, self).__setattr__(attr, val)


class TorchDataLoader(IDataLoader):
    dataset: TorchDataset

    def __init__(
        self,
        loader: _DataLoader,
        *,
        sample_weights: Optional[np.ndarray] = None,
    ):
        if sample_weights is not None:
            raise ValueError(
                "in `DLLoader`, we should introduce `sample_weights` to the original "
                "Pytorch `DataLoader` (by specifying corresponding samplers)"
            )
        super().__init__(sample_weights=sample_weights)
        self.loader = loader
        self.use_numpy = False
        self.dataset = loader.dataset  # type: ignore
        self.sampler_backup = loader.sampler
        self._iterator: Optional[_BaseDataLoaderIter] = None

    def __iter__(self) -> "TorchDataLoader":
        self._iterator = self.loader.__iter__()
        return self

    def __next__(self) -> np_dict_type:
        if self._iterator is None:
            raise StopIteration
        batch = self._iterator.__next__()
        batch = {
            k: to_numpy(v) if isinstance(v, torch.Tensor) else v
            for k, v in batch.items()
        }
        return batch

    def __len__(self) -> int:
        return len(self.loader)

    @property
    def batch_size(self) -> int:  # type: ignore
        return self.loader.batch_size * get_world_size()

    def copy(self) -> "TorchDataLoader":
        dataset = self.dataset
        self.__dict__.pop("dataset")
        copied = super().copy()
        assert isinstance(copied, TorchDataLoader)
        self.dataset = copied.dataset = dataset
        return copied

    def disable_shuffle(self) -> None:
        sampler = SequentialSampler(self.dataset)
        self.loader.sampler = sampler
        if hasattr(self.loader, "batch_sampler"):
            self.loader.batch_sampler.sampler = sampler

    def recover_shuffle(self) -> None:
        self.loader.sampler = self.sampler_backup
        if hasattr(self.loader, "batch_sampler"):
            self.loader.batch_sampler.sampler = self.sampler_backup


@DataConfig.register("torch")
class TorchDataConfig(DataConfig):
    use_distributed_sampler: Optional[bool] = None


@IData.register("torch")
class TorchData(IData):
    """
    A thin wrapper for general Datasets, utilizing the `DataLoader` system from pytorch.

    Examples
    --------
    >>> train_dataset = ...
    >>> valid_dataset = None / ...
    >>> data = cflearn.TorchData.init(config, processor_config)
    >>> data.fit(train_dataset, x_valid=valid_dataset)

    or

    >>> train_dataset = ...
    >>> valid_dataset = None / ...
    >>> data = cflearn.TorchData.build(
    >>>     train_dataset,
    >>>     valid_dataset,
    >>>     config=config,
    >>>     processor_config=processor_config,
    >>> )

    """

    config: TorchDataConfig
    train_dataset: IDataset
    valid_dataset: Optional[IDataset]

    # inheritance

    @property
    def config_base(self) -> Type[TorchDataConfig]:
        return TorchDataConfig

    @property
    def train_kw(self) -> Dict[str, Any]:
        return dict(
            dataset=self.train_dataset,
            sample_weights=self.train_weights,
            shuffle=self.config.shuffle_train,
            batch_size=self.config.batch_size,
            use_distributed_sampler=self.config.use_distributed_sampler,
        )

    @property
    def valid_kw(self) -> Dict[str, Any]:
        return dict(
            dataset=self.valid_dataset,
            sample_weights=self.valid_weights,
            shuffle=self.config.shuffle_valid,
            batch_size=self.config.valid_batch_size or self.config.batch_size,
            use_distributed_sampler=self.config.use_distributed_sampler,
        )

    def get_loaders(self) -> Tuple[TorchDataLoader, Optional[TorchDataLoader]]:
        if not self.processor.is_ready:
            raise ValueError(
                "`processor` should be ready before calling `initialize`, "
                "did you forget to call the `prepare` method first?"
            )
        if self.bundle is None:
            raise ValueError(
                "`bundle` property is not initialized, "
                "did you forget to call the `fit` method first?"
            )
        self.train_dataset = TorchDataset(self.bundle.x_train, self.processor)  # type: ignore
        train_loader = self._make_loader(self.train_kw)
        if self.bundle.x_valid is None:
            self.valid_dataset = None
            valid_loader = None
        else:
            self.valid_dataset = TorchDataset(self.bundle.x_valid, self.processor)  # type: ignore
            valid_loader = self._make_loader(self.valid_kw)
        return train_loader, valid_loader  # type: ignore

    def _make_loader(self, kw: Dict[str, Any]) -> Optional[TorchDataLoader]:
        dataset = kw.pop("dataset")
        if dataset is None:
            return None
        sw = kw.pop("sample_weights")
        return TorchDataLoader(_DataLoader(dataset, **kw), sample_weights=sw)

    # api

    @classmethod
    def build(
        cls,
        x_train: Union[IDataset, Any],
        x_valid: Optional[Union[IDataset, Any]] = None,
        *,
        config: Optional[TorchDataConfig] = None,
        processor_config: Optional[DataProcessorConfig] = None,
    ) -> "TorchData":
        self: TorchData = cls.init(config, processor_config)
        self.fit(x_train, x_valid=x_valid)
        return self


__all__ = [
    "TorchDataset",
    "TorchDataLoader",
    "TorchDataConfig",
    "TorchData",
]
